"""Input parsing, output information and a class to store the options for
clustering."""

from openbabel import pybel
from openbabel import openbabel
from scipy.spatial.distance import squareform
import os
import sys
import argparse
import numpy as np
import rmsd
import logging
import importlib.util
from typing import Callable, List, Union
from dataclasses import dataclass
from .utils import get_mol_info

if importlib.util.find_spec("qml"):
    has_qml = True
else:
    has_qml = False


@dataclass
class ClustOptions:
    """Class to store the options for clustering."""

    trajfile: str = None
    min_rmsd: float = None
    n_workers: int = None
    method: str = None
    reorder_alg_name: str = None
    reorder_alg: Callable[
        [np.ndarray, np.ndarray, np.ndarray, np.ndarray], np.ndarray
    ] = None
    out_conf_fmt: str = None
    reorder: bool = None
    exclusions: bool = None
    no_hydrogen: bool = None
    input_distmat: bool = None
    save_confs: bool = None
    plot: bool = None
    opt_order: bool = None
    overwrite: bool = None
    final_kabsch: bool = None
    silhouette_score: bool = None
    distmat_name: str = None
    out_clust_name: str = None
    evo_name: str = None
    mds_name: str = None
    dendrogram_name: str = None
    out_conf_name: str = None
    summary_name: str = None
    solute_natoms: int = None
    reorder_excl: np.ndarray = None
    optimal_cut: np.ndarray = None

    def update(self, new: dict) -> None:
        """Update the instance with new values.

        Args:
            new (dict): A dictionary containing the new values to update.

        Returns:
            None
        """
        for key, value in new.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def __str__(self) -> str:
        """Return a string representation of the ClustOptions object.

        Returns:
            str: The string representation of the ClustOptions object.
        """
        return_str = "\nFull command: " + " ".join(sys.argv)

        # main parameters
        return_str += f"\n\nClusterized from trajectory file: {self.trajfile}\n"
        return_str += f"Method: {self.method}\n"
        if self.silhouette_score:
            return_str += "\nUsing silhouette score\n"
            if isinstance(self.optimal_cut, (np.ndarray, list)):
                scut = self.optimal_cut[0]
            elif isinstance(self.optimal_cut, (float, np.float32, np.float64)):
                scut = self.optimal_cut
            else:
                raise ValueError("optimal_cut must be a float or np.ndarray")
            return_str += f"RMSD criterion found by silhouette: {scut}\n"
        else:
            return_str += f"RMSD criterion: {self.min_rmsd}\n"
        return_str += f"Ignoring hydrogens?: {self.no_hydrogen}\n"

        # reordering options
        if self.reorder:
            if self.solute_natoms:
                return_str += "\nUsing solute-solvent reordering\n"
                if self.final_kabsch:
                    return_str += "Using final Kabsch rotation before computing RMSD\n"
                return_str += f"Number of solute atoms: {self.solute_natoms}\n"
            else:
                return_str += "\nReordering all atom at the same time\n"

            return_str += f"Reordering algorithm: {self.reorder_alg_name}\n"

            if self.exclusions:
                exclusions_str = ""
                for val in self.reorder_excl:
                    exclusions_str += "%d " % val
                return_str += f"Atoms that weren't considered in the reordering: {exclusions_str.strip()}\n"

        # write file names
        if self.input_distmat:
            return_str += f"\nDistance matrix was read from: {self.distmat_name}\n"
        else:
            return_str += f"\nDistance matrix was written in: {self.distmat_name}\n"

        return_str += f"The classification of each configuration was written in: {self.out_clust_name}\n"
        if self.save_confs:
            return_str += f"\nThe superposed structures for each cluster were saved at: {self.out_conf_name}\n"
        if self.plot:
            return_str += f"\nPlotting the dendrogram to: {self.dendrogram_name}\n"
            return_str += (
                f"Plotting the multidimensional scaling to 2D to: {self.mds_name}\n"
            )
            return_str += f"Plotting the evolution of the classification with the trajectory to: {self.evo_name}\n\n"

        return return_str


class Logger:
    """Logger class."""

    logformat = "%(asctime)s %(levelname)-8s [%(filename)s:%(lineno)d] <%(funcName)s> %(message)s"
    formatter = logging.Formatter(fmt=logformat)
    logger = logging.getLogger(__name__)

    @classmethod
    def setup(cls, logfile: str) -> None:
        """Set up the logger.

        Args:
            logfile (str): The path to the log file.

        Returns:
            None
        """
        cls.logger.setLevel(logging.INFO)
        cls.fh = logging.FileHandler(logfile)
        cls.fh.setLevel(logging.DEBUG)
        cls.fh.setFormatter(cls.formatter)
        cls.ch = logging.StreamHandler()
        cls.ch.setStream(sys.stdout)
        cls.ch.setLevel(logging.INFO)
        cls.ch.setFormatter(cls.formatter)
        cls.logger.addHandler(cls.fh)
        cls.logger.addHandler(cls.ch)


def check_positive(value: str) -> int:
    """Check if the given value is a positive integer.

    Args:
        value (str): The value to be checked.

    Raises:
        argparse.ArgumentTypeError: If the value is not a positive integer.

    Returns:
        int: The converted positive integer value.
    """
    ivalue = int(value)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError("%s is an invalid positive int value" % value)
    return ivalue


def extant_file(x: str) -> str:
    """Check if a file exists.

    Args:
        x (str): The file path to check.

    Raises:
        argparse.ArgumentTypeError: If the file does not exist.

    Returns:
        str: The input file path if it exists.
    """
    if not os.path.exists(x):
        raise argparse.ArgumentTypeError(f"{x} does not exist")
    return x


def configure_runtime(args_in: List[str]) -> ClustOptions:
    """Configure the runtime based on command line arguments.

    Args:
        args_in (List[str]): The command line arguments.

    Returns:
        argparse.Namespace: The parsed command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="Run a clustering analysis on a trajectory based on the minimal RMSD obtained with a Kabsch superposition."
    )

    parser.add_argument(
        "trajectory_file",
        type=extant_file,
        help="path to the trajectory containing the conformations to be classified",
    )
    parser.add_argument(
        "-f",
        "--force-overwrite",
        dest="force",
        action="store_true",
        help="force overwriting the files",
    )
    parser.add_argument(
        "-np",
        "--nprocesses",
        metavar="NPROCS",
        type=check_positive,
        default=4,
        help="defines the number of processes used to compute the distance matrix and multidimensional representation (default = 4)",
    )
    parser.add_argument(
        "-n",
        "--no-hydrogen",
        action="store_true",
        help="ignore hydrogens when doing the Kabsch superposition and calculating the RMSD",
    )
    parser.add_argument(
        "-p",
        "--plot",
        action="store_true",
        help="enable the multidimensional scaling and dendrogram plot saving the figures in pdf format (filenames use the same basename of the -oc option)",
    )
    parser.add_argument(
        "-m",
        "--method",
        metavar="METHOD",
        default="average",
        help="method used for clustering (see valid methods at https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html) (default: average)",
    )
    parser.add_argument(
        "-cc",
        "--clusters-configurations",
        metavar="EXTENSION",
        help="save superposed configurations for each cluster in EXTENSION format",
    )
    parser.add_argument(
        "-oc",
        "--outputclusters",
        default="clusters.dat",
        metavar="FILE",
        help="file to store the clusters (default: clusters.dat)",
    )
    parser.add_argument(
        "-e",
        "--reorder",
        action="store_true",
        help="reorder atoms of molecules to lower the RMSD",
    )
    parser.add_argument(
        "-eex",
        "--reorder-exclusions",
        metavar="EXCLUDED ATOMS",
        nargs="+",
        type=check_positive,
        help="list of atoms that are ignored when reordering",
    )
    parser.add_argument(
        "--reorder-alg",
        action="store",
        default="hungarian",
        metavar="METHOD",
        help="select which reorder algorithm to use; hungarian (default), brute, distance, qml. Warning: brute is VERY slow)",
    )
    parser.add_argument(
        "-ns",
        "--natoms-solute",
        metavar="NATOMS",
        type=check_positive,
        help="number of solute atoms, to ignore these atoms in the reordering process",
    )
    parser.add_argument(
        "-odl",
        "--optimal-ordering",
        action="store_true",
        help="use optimal ordering in linkage (can be slow for large trees, and only useful for dendrogram visualization)",
    )
    parser.add_argument(
        "--final-kabsch",
        action="store_true",
        help="force a final Kabsch rotation before the RMSD computation (effect only when using -ns and -e)",
    )
    parser.add_argument(
        "--log",
        type=str,
        metavar="FILE",
        default="clusttraj.log",
        help="log file (default: clusttraj.log)",
    )

    rmsd_criterion = parser.add_mutually_exclusive_group(required=True)

    rmsd_criterion.add_argument(
        "-ss",
        "--silhouette-score",
        action="store_true",
        help="use the silhouette to determine the criterion to classify structures",
    )
    rmsd_criterion.add_argument(
        "-rmsd",
        "--min-rmsd",
        type=float,
        help="value of RMSD used to classify structures as similar",
    )

    io_group = parser.add_mutually_exclusive_group()
    io_group.add_argument(
        "-i",
        "--input",
        type=extant_file,
        metavar="FILE",
        help="file containing input distance matrix in condensed form (.npy format)",
    )
    io_group.add_argument(
        "-od",
        "--outputdistmat",
        metavar="FILE",
        default="distmat.npy",
        help="file to store distance matrix in condensed form (default: distmat.npy)",
    )

    if len(sys.argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(args_in)

    # setup the logger
    Logger.setup(args.log)

    # check input consistency
    if args.method not in [
        "single",
        "complete",
        "average",
        "weighted",
        "centroid",
        "median",
        "ward",
    ]:
        parser.error(f"The method you selected with -m ({args.method}) is not valid.")

    if args.reorder_alg not in ["hungarian", "brute", "distance", "qml"]:
        parser.error(
            f"The reorder method you selected with --reorder-method ({args.reorder_alg}) is not valid."
        )

    if (args.reorder_alg == "qml") and (not has_qml):
        parser.error(
            "You must have the development branch of qml installed in order to use it as a reorder method."
        )

    if args.clusters_configurations:
        if args.clusters_configurations not in [
            "acr",
            "adf",
            "adfout",
            "alc",
            "arc",
            "bgf",
            "box",
            "bs",
            "c3d1",
            "c3d2",
            "cac",
            "caccrt",
            "cache",
            "cacint",
            "can",
            "car",
            "ccc",
            "cdx",
            "cdxml",
            "cht",
            "cif",
            "ck",
            "cml",
            "cmlr",
            "com",
            "copy",
            "crk2d",
            "crk3d",
            "csr",
            "cssr",
            "ct",
            "cub",
            "cube",
            "dmol",
            "dx",
            "ent",
            "fa",
            "fasta",
            "fch",
            "fchk",
            "fck",
            "feat",
            "fh",
            "fix",
            "fpt",
            "fract",
            "fs",
            "fsa",
            "g03",
            "g92",
            "g94",
            "g98",
            "gal",
            "gam",
            "gamin",
            "gamout",
            "gau",
            "gjc",
            "gjf",
            "gpr",
            "gr96",
            "gukin",
            "gukout",
            "gzmat",
            "hin",
            "inchi",
            "inp",
            "ins",
            "jin",
            "jout",
            "mcdl",
            "mcif",
            "mdl",
            "ml2",
            "mmcif",
            "mmd",
            "mmod",
            "mol",
            "mol2",
            "molden",
            "molreport",
            "moo",
            "mop",
            "mopcrt",
            "mopin",
            "mopout",
            "mpc",
            "mpd",
            "mpqc",
            "mpqcin",
            "msi",
            "msms",
            "nw",
            "nwo",
            "outmol",
            "pc",
            "pcm",
            "pdb",
            "png",
            "pov",
            "pqr",
            "pqs",
            "prep",
            "qcin",
            "qcout",
            "report",
            "res",
            "rsmi",
            "rxn",
            "sd",
            "sdf",
            "smi",
            "smiles",
            "sy2",
            "t41",
            "tdd",
            "test",
            "therm",
            "tmol",
            "txt",
            "txyz",
            "unixyz",
            "vmol",
            "xed",
            "xml",
            "xyz",
            "yob",
            "zin",
        ]:
            parser.error(
                f"The format you selected to save the clustered superposed configurations ({args.clusters_configurations}) is not valid."
            )

    if args.reorder_exclusions and args.no_hydrogen:
        parser.error(
            "You cannot use --no-hydrogen and reorder exclusions with -eex at the same time"
        )

    if not args.reorder and args.reorder_exclusions:
        logging.warning(
            "The list of atoms to exclude for reordering only makes sense if reordering is enabled. Ignoring the list."
        )

    return parse_args(args)


def parse_args(args: argparse.Namespace) -> ClustOptions:
    """Parse all the information from the argument parser, storing in the
    ClustOptions class.

    Define file names and set the pointers to the correct functions.

    Args:
        args (Namespace): The arguments parsed from the argument parser.

    Returns:
        ClustOptions: An instance of the ClustOptions class with the parsed options.
    """
    basenameout = os.path.splitext(args.outputclusters)[0]

    options_dict = {
        "solute_natoms": args.natoms_solute,
        "reorder_excl": np.asarray([x - 1 for x in args.reorder_exclusions], np.int32)
        if args.reorder_exclusions
        else np.asarray([], np.int32),
        "exclusions": bool(args.reorder_exclusions),
        "reorder_alg_name": args.reorder_alg,
        "reorder_alg": None,
        "reorder": bool(args.reorder),
        "input_distmat": bool(args.input),
        "distmat_name": args.outputdistmat if not args.input else args.input,
        "out_clust_name": args.outputclusters,
        "summary_name": basenameout + ".out",
        "save_confs": bool(args.clusters_configurations),
        "out_conf_name": basenameout + "_confs"
        if args.clusters_configurations
        else None,
        "out_conf_fmt": args.clusters_configurations
        if args.clusters_configurations
        else None,
        "plot": bool(args.plot),
        "evo_name": basenameout + "_evo.pdf" if args.plot else None,
        "dendrogram_name": basenameout + "_dendrogram.pdf" if args.plot else None,
        "mds_name": basenameout + ".pdf" if args.plot else None,
        "trajfile": args.trajectory_file,
        "min_rmsd": args.min_rmsd,
        "method": args.method,
        "n_workers": args.nprocesses,
        "no_hydrogen": args.no_hydrogen,
        "opt_order": args.optimal_ordering,
        "overwrite": args.force,
        "final_kabsch": args.final_kabsch,
        "silhouette_score": args.silhouette_score,
    }

    if args.reorder:
        if args.reorder_alg == "hungarian":
            options_dict["reorder_alg"] = rmsd.reorder_hungarian
        elif args.reorder_alg == "distance":
            options_dict["reorder_alg"] = rmsd.reorder_distance
        elif args.reorder_alg == "brute":
            options_dict["reorder_alg"] = rmsd.reorder_brute
        elif args.reorder_alg == "qml":
            options_dict["reorder_alg"] = rmsd.reorder_similarity

    if not args.input and os.path.exists(args.outputdistmat) and not args.force:
        raise FileExistsError(
            f"File {args.outputdistmat} already exists, specify a new filename with the -od command option or use the -f option. If you are trying to read the distance matrix from a file, use the -i option."
        )

    if os.path.exists(args.outputclusters) and not args.force:
        raise FileExistsError(
            f"File {args.outputclusters} already exists, specify a new filename with the -oc command option or use the -f option."
        )

    return ClustOptions(**options_dict)


def save_clusters_config(
    trajfile: str,
    clusters: np.ndarray,
    distmat: np.ndarray,
    noh: bool,
    reorder: Union[
        Callable[[np.ndarray, np.ndarray, np.ndarray, np.ndarray], np.ndarray], None
    ],
    nsatoms: int,
    outbasename: str,
    outfmt: str,
    reorderexcl: np.ndarray,
    final_kabsch: bool,
    overwrite: bool,
) -> None:
    """Save best superpositioned configurations for each cluster. First
    configuration is the medoid.

    Args:
        trajfile: The trajectory file path.
        clusters: An array containing cluster labels.
        distmat: The distance matrix.
        noh: Flag indicating whether to exclude hydrogen atoms.
        reorder (Union[Callable[[np.ndarray, np.ndarray, np.ndarray, np.ndarray], np.ndarray], None]):
            A function to reorder the atoms, if necessary.
        nsatoms: The number of atoms in the solute.
        outbasename: The base name for the output files.
        outfmt: The output file format.
        reorderexcl: An array of atom indices to exclude during reordering.
        final_kabsch: Flag indicating whether to perform a final Kabsch rotation.
        overwrite: Flag indicating whether to overwrite existing output files.

    Returns:
        None
    """
    # complete distance matrix
    sqdistmat = squareform(distmat)

    for cnum in range(1, max(clusters) + 1):
        # create object to output the configurations
        outfile = pybel.Outputfile(
            outfmt, outbasename + "_" + str(cnum) + "." + outfmt, overwrite=overwrite
        )

        # creates mask with True only for the members of cluster number cnum
        mask = np.array([1 if i == cnum else 0 for i in clusters], dtype=bool)

        # gets the member with smallest sum of distances from the submatrix
        idx = np.argmin(sum(sqdistmat[:, mask][mask, :]))

        # get list with the members of this cluster only and store medoid
        sublist = [num for (num, cluster) in enumerate(clusters) if cluster == cnum]
        medoid = sublist[idx]

        # get the medoid coordinates
        for idx, mol in enumerate(
            pybel.readfile(os.path.splitext(trajfile)[1][1:], trajfile)
        ):
            if idx != medoid:
                continue

            # medoid coordinates
            tnatoms = len(mol.atoms)
            q_atoms, q_all = get_mol_info(mol)

            # get the number of non hydrogen atoms in the solute to subtract if needed
            natoms = nsatoms
            if noh:
                natoms = len(np.where(q_atoms[:nsatoms] != 1)[0])

            if nsatoms:
                if noh:
                    not_hydrogens = np.where(q_atoms != 1)
                    Q = np.copy(q_all[not_hydrogens])
                    Qa = np.copy(q_atoms[not_hydrogens])
                else:
                    Q = np.copy(q_all)
                    Qa = np.copy(q_atoms)

                qcenter = rmsd.centroid(Q[:natoms])
            elif noh:
                not_hydrogens = np.where(q_atoms != 1)
                Q = np.copy(q_all[not_hydrogens])
                qcenter = rmsd.centroid(Q)
                Qa = np.copy(q_atoms[not_hydrogens])
            else:
                Q = np.copy(q_all)
                qcenter = rmsd.centroid(Q)
                Qa = np.copy(q_atoms)

            # center the coordinates at the origin
            Q -= qcenter

            # write medoid configuration to file (molstring is a xyz string used to generate de pybel mol)
            molstring = str(tnatoms) + "\n" + mol.title.rstrip() + "\n"
            for i, coords in enumerate(q_all - qcenter):
                molstring += (
                    openbabel.GetSymbol(int(q_atoms[i]))
                    + "\t"
                    + str(coords[0])
                    + "\t"
                    + str(coords[1])
                    + "\t"
                    + str(coords[2])
                    + "\n"
                )
            rmol = pybel.readstring("xyz", molstring)
            outfile.write(rmol)

            break

        # rotate all the cluster members into the medoid and print them to the .xyz file
        for idx, mol in enumerate(
            pybel.readfile(os.path.splitext(trajfile)[1][1:], trajfile)
        ):
            if not mask[idx] or idx == medoid:
                continue

            # config coordinates
            p_atoms, p_all = get_mol_info(mol)

            if nsatoms:
                if noh:
                    not_hydrogens = np.where(p_atoms != 1)
                    P = np.copy(p_all[not_hydrogens])
                    Pa = np.copy(p_atoms[not_hydrogens])
                else:
                    P = np.copy(p_all)
                    Pa = np.copy(p_atoms)

                pcenter = rmsd.centroid(P[:natoms])
            elif noh:
                not_hydrogens = np.where(p_atoms != 1)
                P = np.copy(p_all[not_hydrogens])
                pcenter = rmsd.centroid(P)
                Pa = np.copy(p_atoms[not_hydrogens])
            else:
                P = np.copy(p_all)
                pcenter = rmsd.centroid(P)
                Pa = np.copy(p_atoms)

            # center the coordinates at the origin
            P -= pcenter
            p_all -= pcenter

            # generate rotation to superpose the solute configuration
            if nsatoms:
                # center the coordinates at the solute
                P -= rmsd.centroid(Q[:natoms])

                # try to improve atom matching by performing Kabsch
                # generate a rotation considering only the solute atoms
                U = rmsd.kabsch(P[:natoms], Q[:natoms])

                # rotate the whole system with this rotation
                P = np.dot(P, U)
                p_all = np.dot(p_all, U)

                # reorder solute atoms
                if reorder:
                    # find the solute atoms that are not excluded
                    soluexcl = np.where(reorderexcl < natoms)
                    soluteview = np.delete(np.arange(natoms), reorderexcl[soluexcl])
                    Pview = P[soluteview]
                    Paview = Pa[soluteview]

                    # reorder just these atoms
                    prr = reorder(Qa[soluteview], Paview, Q[soluteview], Pview)
                    Pview = Pview[prr]
                    Paview = Paview[prr]

                    # build the total molecule reordering just these atoms
                    whereins = np.where(
                        np.isin(np.arange(natoms), reorderexcl[soluexcl]) is True
                    )
                    Psolu = np.insert(
                        Pview,
                        [x - whereins[0].tolist().index(x) for x in whereins[0]],
                        P[reorderexcl[soluexcl]],
                        axis=0,
                    )
                    Pasolu = np.insert(
                        Paview,
                        [x - whereins[0].tolist().index(x) for x in whereins[0]],
                        Pa[reorderexcl[soluexcl]],
                        axis=0,
                    )

                    P = np.concatenate((Psolu, P[np.arange(len(P) - natoms) + natoms]))
                    Pa = np.concatenate(
                        (Pasolu, Pa[np.arange(len(Pa) - natoms) + natoms])
                    )

                    # generate a rotation considering only the solute atoms
                    U = rmsd.kabsch(P[:natoms], Q[:natoms])

                    # rotate the whole system with this rotation
                    P = np.dot(P, U)
                    p_all = np.dot(p_all, U)

                    # consider only the solvent atoms in the reorder (without exclusions)
                    # solvexcl = np.where(reorderexcl >= natoms)
                    # solvview = np.delete(np.arange(natoms, len(P)), reorderexcl[solvexcl])
                    # Pview = P[solvview]
                    # Paview = Pa[solvview]

                    # # reorder just these atoms
                    # prr = reorder(Qa[solvview], Paview, Q[solvview], Pview)
                    # Pview = Pview[prr]
                    # Paview = Paview[prr]

                    # # build the total molecule with the reordered atoms
                    # whereins = np.where(
                    #     np.isin(np.arange(natoms, len(P)), reorderexcl[solvexcl]) == True
                    # )
                    # Psolv = np.insert(
                    #     Pview,
                    #     [x - whereins[0].tolist().index(x) for x in whereins[0]],
                    #     P[reorderexcl[solvexcl]],
                    #     axis=0,
                    # )
                    # Pasolv = np.insert(
                    #     Paview,
                    #     [x - whereins[0].tolist().index(x) for x in whereins[0]],
                    #     Pa[reorderexcl[solvexcl]],
                    #     axis=0,
                    # )

                    # Pr = np.concatenate((P[:natoms], Psolv))
                    # Pra = np.concatenate((Pa[:natoms], Pasolv))
            else:
                # Kabsch rotation
                U = rmsd.kabsch(P, Q)
                P = np.dot(P, U)
                p_all = np.dot(p_all, U)

            # reorder the solvent atoms separately
            if reorder:
                # get the view without the excluded atoms
                view = np.delete(np.arange(len(P)), reorderexcl)
                Pview = P[view]
                Paview = Pa[view]

                prr = reorder(Qa[view], Paview, Q[view], Pview)
                Pview = Pview[prr]

                # build the total molecule with the reordered atoms
                whereins = np.where(np.isin(np.arange(len(P)), reorderexcl) is True)
                Pr = np.insert(
                    Pview,
                    [x - whereins[0].tolist().index(x) for x in whereins[0]],
                    P[reorderexcl],
                    axis=0,
                )
            else:
                Pr = P

            if nsatoms and reorder and not final_kabsch:
                # rotate whole configuration (considering hydrogens even with noh)
                U = rmsd.kabsch(Pr, Q)
                p_all = np.dot(p_all, U)

            # write rotated configuration to file (molstring is a xyz string used to generate de pybel mol)
            molstring = str(tnatoms) + "\n" + mol.title.rstrip() + "\n"
            for i, coords in enumerate(p_all):
                molstring += (
                    openbabel.GetSymbol(int(p_atoms[i]))
                    + "\t"
                    + str(coords[0])
                    + "\t"
                    + str(coords[1])
                    + "\t"
                    + str(coords[2])
                    + "\n"
                )
            rmol = pybel.readstring("xyz", molstring)
            outfile.write(rmol)

        # closes the file for the cnum cluster
        outfile.close()



import pytest
import numpy as np
from clusttraj.utils import get_mol_coords, get_mol_info


def test_get_mol_coords(pybel_mol):
    coords = get_mol_coords(pybel_mol)

    expected_coords = np.array(
        [[0.0, 0.0, 0.11779], [0.0, 0.75545, -0.47116], [0.0, -0.75545, -0.47116]]
    )
    assert coords == pytest.approx(expected_coords, abs=1e-8)


def test_get_mol_info(pybel_mol):
    atomicnums, coords = get_mol_info(pybel_mol)

    expected_atomicnums = np.array([8, 1, 1])
    expected_coords = np.array(
        [[0.0, 0.0, 0.11779], [0.0, 0.75545, -0.47116], [0.0, -0.75545, -0.47116]]
    )

    assert atomicnums == pytest.approx(expected_atomicnums, abs=1e-8)
    assert coords == pytest.approx(expected_coords, abs=1e-8)



"""Functions to perform clustering based on the distance matrix."""

import scipy.cluster.hierarchy as hcl
from scipy.spatial.distance import squareform
from sklearn import metrics
import numpy as np
from typing import Tuple
from .io import ClustOptions, Logger


def classify_structures_silhouette(
    clust_opt: ClustOptions, distmat: np.ndarray, dstep: float = 0.1
) -> Tuple[np.ndarray, np.ndarray]:
    """Find the optimal threshold following the silhouette score metric and
    perform the classification.

    Args:
        clust_opt: The clustering options.
        distmat: The distance matrix.
        dstep (float, optional): Interval between threshold values, defaults to 0.1

    Returns:
        A tuple containing the linkage matrix and the clusters.
    """

    # linkage
    Logger.logger.info(
        f"Clustering using '{clust_opt.method}' method to join the clusters\n"
    )
    Z = hcl.linkage(distmat, clust_opt.method, optimal_ordering=clust_opt.opt_order)

    # Initialize the score and threshold
    ss_opt = np.float64()
    t_opt = np.array([])
    labels_opt = np.array([])

    # Get the range of threshold values
    t_range = np.arange(min(Z[:, 2]), max(Z[:, 2]), dstep)

    for t in t_range:
        # Create an array with cluster labels
        hcl_labels = hcl.fcluster(Z, t=t, criterion="distance")

        # Compute the silhouette score
        ss = metrics.silhouette_score(
            squareform(distmat), hcl_labels, metric="precomputed"
        )

        # Check for degeneracy for the optimal threshold value
        if np.any(ss == ss_opt):
            ss_opt = ss
            t_opt = np.append(t_opt, t)
            labels_opt = np.vstack((labels_opt, hcl_labels))

        # Update the values to the highest silhouette score
        if np.all(ss > ss_opt):
            ss_opt = ss
            t_opt = t
            labels_opt = hcl_labels

    Logger.logger.info(f"Highest silhouette score: {ss_opt}\n")

    if t_opt.size > 1:
        t_opt_str = ", ".join([str(t) for t in t_opt])
        Logger.logger.info(
            f"The following RMSD threshold values yielded the same optimial silhouette score: {t_opt_str}\n"
        )
        Logger.logger.info(f"The smallest RMSD of {t_opt[0]} has been adopted\n")
        clusters = labels_opt[0]
    else:
        Logger.logger.info(f"Optimal RMSD threshold value: {t_opt}\n")
        clusters = labels_opt

    clust_opt.update({"optimal_cut": t_opt})

    Logger.logger.info(
        f"Saving clustering classification to {clust_opt.out_clust_name}\n"
    )
    np.savetxt(clust_opt.out_clust_name, clusters, fmt="%d")

    return Z, clusters


def classify_structures(
    clust_opt: ClustOptions, distmat: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Classify structures based on clustering options and distance matrix.

    Args:
        clust_opt: The clustering options.
        distmat: The distance matrix.

    Returns:
        A tuple containing the linkage matrix and the clusters.
    """
    # linkage
    Logger.logger.info(
        f"Clustering using '{clust_opt.method}' method to join the clusters\n"
    )
    Z = hcl.linkage(distmat, clust_opt.method, optimal_ordering=clust_opt.opt_order)

    # build the clusters
    clusters = hcl.fcluster(Z, clust_opt.min_rmsd, criterion="distance")

    Logger.logger.info(
        f"Saving clustering classification to {clust_opt.out_clust_name}\n"
    )
    np.savetxt(clust_opt.out_clust_name, clusters, fmt="%d")

    return Z, clusters



"""Functions to compute the distance matrix based on the provided
trajectory."""

from openbabel import pybel
import numpy as np
import rmsd
import os
import multiprocessing
import itertools
from .io import ClustOptions, Logger
from .utils import get_mol_info
from typing import List, Union, Callable


def get_distmat(clust_opt: ClustOptions) -> np.ndarray:
    """Calculate or read a condensed distance matrix based on the given
    clustering options.

    Args:
        clust_opt (ClustOptions): The clustering options.

    Returns:
        np.ndarray: The condensed distance matrix.
    """
    # check if distance matrix will be read from input or calculated
    # if a file is specified, read it (TODO: check if the matrix makes sense)
    if clust_opt.input_distmat:
        Logger.logger.info(
            f"Reading condensed distance matrix from {clust_opt.distmat_name}\n"
        )
        distmat = np.load(clust_opt.distmat_name)
    # build a distance matrix already in the condensed form
    else:
        Logger.logger.info(
            f"Calculating distance matrix using {clust_opt.n_workers} threads\n"
        )
        distmat = build_distance_matrix(clust_opt)
        Logger.logger.info(
            f"Saving condensed distance matrix to {clust_opt.distmat_name}\n"
        )
        np.save(clust_opt.distmat_name, distmat)

    return distmat


def build_distance_matrix(clust_opt: ClustOptions) -> np.ndarray:
    """Compute the distance matrix.

    Args:
        clust_opt (ClustOptions): The options for clustering.

    Returns:
        np.ndarray: The computed distance matrix.
    """
    # create iterator containing information to compute a line of the distance matrix
    inputiterator = zip(
        itertools.count(),
        map(
            lambda x: get_mol_info(x),
            pybel.readfile(
                os.path.splitext(clust_opt.trajfile)[1][1:], clust_opt.trajfile
            ),
        ),
        itertools.repeat(clust_opt.trajfile),
        itertools.repeat(clust_opt.no_hydrogen),
        itertools.repeat(clust_opt.reorder_alg),
        itertools.repeat(clust_opt.solute_natoms),
        itertools.repeat(clust_opt.reorder_excl),
        itertools.repeat(clust_opt.final_kabsch),
    )

    # create the pool with nprocs processes to compute the distance matrix in parallel
    p = multiprocessing.Pool(processes=clust_opt.n_workers)

    # build the distance matrix in parallel
    ldistmat = p.starmap(compute_distmat_line, inputiterator)

    return np.asarray([x for n in ldistmat if len(n) > 0 for x in n])


def compute_distmat_line(
    idx1: int,
    q_info: tuple,
    trajfile: str,
    noh: bool,
    reorder: Union[
        Callable[[np.ndarray, np.ndarray, np.ndarray, np.ndarray], np.ndarray], None
    ],
    nsatoms: int,
    reorderexcl: np.ndarray,
    final_kabsch: bool,
) -> List[float]:
    """Compute the distance between molecule idx1 and molecules with idx2 >
    idx1.

    Args:
        idx1 (int): The index of the first molecule.
        q_info (tuple): Tuple containing the atom and all information of the first molecule.
        trajfile (str): The path to the trajectory file.
        noh (bool): Whether to consider hydrogen atoms or not.
        reorder (Union[Callable[[np.ndarray, np.ndarray, np.ndarray, np.ndarray], np.ndarray], None]):
            A function to reorder the atoms, if necessary.
        nsatoms (int): The number of atoms in the solute.
        reorderexcl (np.ndarray): The array defining the excluded atoms during reordering.
        final_kabsch (bool): Whether to perform the final Kabsch rotation or not.

    Returns:
        List[float]: The distance matrix.
    """  # noqa: E501
    # unpack q_info tuple
    q_atoms, q_all = q_info

    # get the number of non hydrogen atoms in the solute to subtract if needed
    natoms = nsatoms
    if noh:
        natoms = len(np.where(q_atoms[:nsatoms] != 1)[0])

    # initialize distance matrix
    distmat = []

    for idx2, mol2 in enumerate(
        pybel.readfile(os.path.splitext(trajfile)[1][1:], trajfile)
    ):
        # skip if it's not an element from the superior diagonal matrix
        if idx1 >= idx2:
            continue

        # arrays for second molecule
        p_atoms, p_all = get_mol_info(mol2)

        # consider the H or not consider depending on option
        if nsatoms:
            if noh:
                not_hydrogensP = np.where(p_atoms != 1)
                not_hydrogensQ = np.where(q_atoms != 1)
                P = p_all[not_hydrogensP]
                Q = q_all[not_hydrogensQ]
                Pa = p_atoms[not_hydrogensP]
                Qa = q_atoms[not_hydrogensQ]
            else:
                P = p_all
                Q = q_all
                Pa = p_atoms
                Qa = q_atoms

            pcenter = rmsd.centroid(P[:natoms])
            qcenter = rmsd.centroid(Q[:natoms])
        elif noh:
            not_hydrogensP = np.where(p_atoms != 1)
            not_hydrogensQ = np.where(q_atoms != 1)
            P = p_all[not_hydrogensP]
            Q = q_all[not_hydrogensQ]
            Pa = p_atoms[not_hydrogensP]
            Qa = q_atoms[not_hydrogensQ]
            pcenter = rmsd.centroid(P)
            qcenter = rmsd.centroid(Q)
        else:
            P = p_all
            Q = q_all
            Pa = p_atoms
            Qa = q_atoms
            pcenter = rmsd.centroid(P)
            qcenter = rmsd.centroid(Q)

        # center the coordinates at the origin
        P -= pcenter
        Q -= qcenter

        # generate rotation to superpose the solute configuration
        if nsatoms:
            # center the coordinates at the solute
            P -= rmsd.centroid(Q[:natoms])

            # try to improve atom matching by performing Kabsch
            # generate a rotation considering only the solute atoms without reordering
            U = rmsd.kabsch(P[:natoms], Q[:natoms])

            # rotate the whole system with this rotation
            P = np.dot(P, U)

            # reorder solute atoms
            if reorder:
                # find the solute atoms that are not excluded
                soluexcl = np.where(reorderexcl < natoms)
                soluteview = np.delete(np.arange(natoms), reorderexcl[soluexcl])
                Pview = P[soluteview]
                Paview = Pa[soluteview]

                # reorder just these atoms
                prr = reorder(Qa[soluteview], Paview, Q[soluteview], Pview)
                Pview = Pview[prr]
                Paview = Paview[prr]

                # build the total structure reordering just these atoms
                whereins = np.where(
                    np.isin(np.arange(natoms), reorderexcl[soluexcl]) is True
                )
                Psolu = np.insert(
                    Pview,
                    [x - whereins[0].tolist().index(x) for x in whereins[0]],
                    P[reorderexcl[soluexcl]],
                    axis=0,
                )
                Pasolu = np.insert(
                    Paview,
                    [x - whereins[0].tolist().index(x) for x in whereins[0]],
                    Pa[reorderexcl[soluexcl]],
                    axis=0,
                )

                P = np.concatenate((Psolu, P[np.arange(len(P) - natoms) + natoms]))
                Pa = np.concatenate((Pasolu, Pa[np.arange(len(Pa) - natoms) + natoms]))

                # generate a rotation considering the reordered solute atoms
                U = rmsd.kabsch(P[:natoms], Q[:natoms])

                # rotate the whole system with this rotation
                P = np.dot(P, U)

                # consider only the solvent atoms in the reorder (without exclusions)
                # solvexcl = np.where(reorderexcl >= natoms)
                # solvview = np.delete(np.arange(natoms, len(P)), reorderexcl[solvexcl])
                # Pview = P[solvview]
                # Paview = Pa[solvview]

                # # reorder just these atoms
                # prr = reorder(Qa[solvview], Paview, Q[solvview], Pview)
                # Pview = Pview[prr]
                # Paview = Paview[prr]

                # # build the total molecule with the reordered atoms
                # whereins = np.where(
                #     np.isin(np.arange(natoms, len(P)), reorderexcl[solvexcl]) == True
                # )
                # Psolv = np.insert(
                #     Pview,
                #     [x - whereins[0].tolist().index(x) for x in whereins[0]],
                #     P[reorderexcl[solvexcl]],
                #     axis=0,
                # )
                # Pasolv = np.insert(
                #     Paview,
                #     [x - whereins[0].tolist().index(x) for x in whereins[0]],
                #     Pa[reorderexcl[solvexcl]],
                #     axis=0,
                # )

                # Pr = np.concatenate((P[:natoms], Psolv))
                # Pra = np.concatenate((Pa[:natoms], Pasolv))
        else:
            # Kabsch rotation
            U = rmsd.kabsch(P, Q)
            P = np.dot(P, U)

        # reorder the solvent atoms separately
        if reorder:
            # get the view without the excluded atoms
            view = np.delete(np.arange(len(P)), reorderexcl)
            Pview = P[view]
            Paview = Pa[view]

            prr = reorder(Qa[view], Paview, Q[view], Pview)
            Pview = Pview[prr]

            # build the total molecule with the reordered atoms
            whereins = np.where(np.isin(np.arange(len(P)), reorderexcl) is True)
            Pr = np.insert(
                Pview,
                [x - whereins[0].tolist().index(x) for x in whereins[0]],
                P[reorderexcl],
                axis=0,
            )

        else:
            Pr = P

        # for solute solvent alignement, compute RMSD without Kabsch
        if nsatoms and reorder and not final_kabsch:
            distmat.append(rmsd.rmsd(Pr, Q))
        else:
            distmat.append(rmsd.kabsch_rmsd(Pr, Q))

    return distmat



import os
from clusttraj.main import main


def test_main(tmp_path):
    main(
        [
            "test/ref/testtraj.xyz",
            "--min-rmsd",
            "1.0",
            "-np",
            "1",
            "-i",
            "test/ref/test_distmat.npy",
            "-oc",
            os.path.join(tmp_path, "clusters.dat"),
        ]
    )

    assert os.path.exists(os.path.join(tmp_path, "clusters.dat"))
    assert os.path.exists(os.path.join(tmp_path, "clusters.out"))



"""Main entry point for clusttraj.

Can be called from command line or from an external library given a list
of arguments.
"""

import sys
import numpy as np
from typing import List
from .io import Logger, configure_runtime, save_clusters_config
from .distmat import get_distmat
from .plot import plot_clust_evo, plot_dendrogram, plot_mds
from .classify import classify_structures, classify_structures_silhouette


def main(args: List[str] = None) -> None:
    """Main function that performs clustering and generates output.

    Args:
        args (List): List of command-line arguments. Defaults to None.

    Returns:
        None
    """
    if args is None:
        args = sys.argv[1:]

    # get ClustOptions class with parsed arguments
    clust_opt = configure_runtime(args)

    # get the distance matrix
    distmat = get_distmat(clust_opt)

    # perform the clustering
    if clust_opt.silhouette_score:
        Z, clusters = classify_structures_silhouette(clust_opt, distmat)
    else:
        Z, clusters = classify_structures(clust_opt, distmat)

    # get the elements closest to the centroid (see https://stackoverflow.com/a/39870085/3254658)
    if clust_opt.save_confs:
        outconf = clust_opt.out_conf_name + "_*." + clust_opt.out_conf_fmt
        Logger.logger.info(
            f"Writing superposed configurations per cluster to files {outconf}\n"
        )
        save_clusters_config(
            clust_opt.trajfile,
            clusters,
            distmat,
            clust_opt.no_hydrogen,
            clust_opt.reorder_alg,
            clust_opt.solute_natoms,
            clust_opt.out_conf_name,
            clust_opt.out_conf_fmt,
            clust_opt.reorder_excl,
            clust_opt.final_kabsch,
            clust_opt.overwrite,
        )

    # generate plots
    if clust_opt.plot:
        plot_clust_evo(clust_opt, clusters)

        plot_dendrogram(clust_opt, Z)

        plot_mds(clust_opt, clusters, distmat)

    # print the cluster sizes
    outclust_str = f"A total {len(clusters)} snapshots were read and {max(clusters)} cluster(s) was(were) found.\n"
    outclust_str += "The cluster sizes are:\nCluster\tSize\n"

    labels, sizes = np.unique(clusters, return_counts=True)
    for label, size in zip(labels, sizes):
        outclust_str += f"{label}\t{size}\n"
    Logger.logger.info(outclust_str)

    # save summary
    with open(clust_opt.summary_name, "w") as f:
        f.write(str(clust_opt))
        f.write(outclust_str)


if __name__ == "__main__":
    main()



import pytest
from clusttraj.utils import get_mol_info
from clusttraj.distmat import build_distance_matrix, get_distmat, compute_distmat_line
from clusttraj.io import ClustOptions


def test_get_distmat(options_dict, test_distmat):
    distmat = get_distmat(ClustOptions(**options_dict))
    assert len(distmat) == 3
    assert distmat == pytest.approx(test_distmat, abs=1e-8)

    options_dict["input_distmat"] = True
    distmat = get_distmat(ClustOptions(**options_dict))
    assert len(distmat) == 3
    assert distmat == pytest.approx(test_distmat, abs=1e-8)


def test_compute_distmat_line(options_dict, clust_opt, first_conf_traj, test_distmat):
    print(options_dict)
    line = compute_distmat_line(
        0,
        get_mol_info(first_conf_traj),
        clust_opt.trajfile,
        clust_opt.no_hydrogen,
        None,
        clust_opt.solute_natoms,
        clust_opt.reorder_excl,
        clust_opt.final_kabsch,
    )

    print(line)
    print(test_distmat)
    assert len(line) == 2
    assert line[0] == pytest.approx(test_distmat[0], abs=1e-8)
    assert line[1] == pytest.approx(test_distmat[1], abs=1e-8)


def test_build_distance_matrix(clust_opt, test_distmat):
    distmat = build_distance_matrix(clust_opt)
    assert len(distmat) == 3
    assert distmat[0] == pytest.approx(test_distmat[0], abs=1e-8)
    assert distmat[1] == pytest.approx(test_distmat[1], abs=1e-8)
    assert distmat[2] == pytest.approx(test_distmat[2], abs=1e-8)



import pytest
import argparse
import os
import numpy as np
import rmsd
from clusttraj.io import (
    ClustOptions,
    check_positive,
    extant_file,
    configure_runtime,
    parse_args,
    save_clusters_config,
)


def test_ClustOptions(options_dict):
    clust_opt = ClustOptions(**options_dict)

    assert os.path.basename(clust_opt.out_clust_name) == "clusters.dat"
    assert len(clust_opt.reorder_excl) == 0
    assert isinstance(clust_opt.reorder_excl, np.ndarray)
    assert clust_opt.exclusions is False
    assert clust_opt.reorder_alg_name == "hungarian"
    assert clust_opt.reorder_alg is None
    assert clust_opt.reorder is False
    assert clust_opt.input_distmat is False
    assert clust_opt.distmat_name == "test/ref/test_distmat.npy"
    assert os.path.basename(clust_opt.summary_name) == "clusters.out"
    assert os.path.basename(clust_opt.evo_name) == "clusters_evo.pdf"
    assert os.path.basename(clust_opt.dendrogram_name) == "clusters_dendrogram.pdf"
    assert os.path.basename(clust_opt.out_conf_name) == "clusters_confs"
    assert clust_opt.out_conf_fmt == "xyz"
    assert os.path.basename(clust_opt.mds_name) == "clusters.pdf"
    assert clust_opt.save_confs is False
    assert clust_opt.plot is False
    assert clust_opt.trajfile == "test/ref/testtraj.xyz"
    assert clust_opt.min_rmsd == 1.0
    assert clust_opt.method == "ward"
    assert clust_opt.n_workers == 1
    assert clust_opt.no_hydrogen is True
    assert clust_opt.opt_order is False
    assert clust_opt.solute_natoms == 17
    assert clust_opt.overwrite is True
    assert clust_opt.final_kabsch is False
    assert clust_opt.silhouette_score is False


def test_check_positive():
    assert check_positive("1") == 1

    with pytest.raises(ValueError):
        check_positive("a")

    with pytest.raises(argparse.ArgumentTypeError):
        check_positive("-1")


def test_extant_file():
    assert extant_file("test/ref/testtraj.xyz") == "test/ref/testtraj.xyz"

    with pytest.raises(argparse.ArgumentTypeError):
        extant_file("somewhere/nonexistent.file")


def test_parse_args():
    args = argparse.Namespace(
        natoms_solute=10,
        reorder_exclusions=[1, 2, 3],
        reorder_alg="hungarian",
        reorder=False,
        input=True,
        outputdistmat="distmat.npy",
        outputclusters="clusters.dat",
        clusters_configurations=True,
        plot=True,
        trajectory_file="trajectory.xyz",
        min_rmsd=0.1,
        method="ward",
        nprocesses=4,
        no_hydrogen=True,
        optimal_ordering=True,
        force=True,
        final_kabsch=True,
        silhouette_score=False,
    )
    clust_opt = parse_args(args)

    assert isinstance(clust_opt, ClustOptions)
    assert clust_opt.reorder_alg is None

    args = argparse.Namespace(
        natoms_solute=10,
        reorder_exclusions=[1, 2, 3],
        reorder_alg="hungarian",
        reorder=True,
        input=True,
        outputdistmat="distmat.npy",
        outputclusters="clusters.dat",
        clusters_configurations=True,
        plot=True,
        trajectory_file="trajectory.xyz",
        min_rmsd=0.1,
        method="ward",
        nprocesses=4,
        no_hydrogen=True,
        optimal_ordering=True,
        force=True,
        final_kabsch=True,
        silhouette_score=False,
    )
    clust_opt = parse_args(args)

    assert clust_opt.reorder_alg == rmsd.reorder_hungarian


def test_configure_runtime(caplog):
    clust_opt = configure_runtime(
        ["test/ref/testtraj.xyz", "--min-rmsd", "1.0", "-np", "1"]
    )

    assert clust_opt.trajfile == "test/ref/testtraj.xyz"
    assert clust_opt.min_rmsd == pytest.approx(1.0, abs=1e-8)
    assert clust_opt.n_workers == 1
    assert clust_opt.out_clust_name == "clusters.dat"

    with pytest.raises(SystemExit):
        clust_opt = configure_runtime(
            ["test/ref/testtraj.xyz", "--min-rmsd", "1.0", "-m", "nonexistent-method"]
        )

    with pytest.raises(SystemExit):
        clust_opt = configure_runtime(
            [
                "test/ref/testtraj.xyz",
                "--min-rmsd",
                "1.0",
                "--reorder-alg",
                "nonexistent-method",
            ]
        )

    with pytest.raises(SystemExit):
        clust_opt = configure_runtime(
            [
                "test/ref/testtraj.xyz",
                "--min-rmsd",
                "1.0",
                "-cc",
                "nonexistent-extension",
            ]
        )

    with pytest.raises(SystemExit):
        clust_opt = configure_runtime(
            ["test/ref/testtraj.xyz", "--min-rmsd", "1.0", "-n", "-eex", "1"]
        )


def test_save_clusters_config(clust_opt, clusters_seq, test_distmat):
    save_clusters_config(
        clust_opt.trajfile,
        clusters_seq,
        test_distmat,
        clust_opt.no_hydrogen,
        clust_opt.reorder_alg,
        clust_opt.solute_natoms,
        clust_opt.out_conf_name,
        clust_opt.out_conf_fmt,
        clust_opt.reorder_excl,
        clust_opt.final_kabsch,
        clust_opt.overwrite,
    )

    assert os.path.exists(clust_opt.out_conf_name + "_1." + clust_opt.out_conf_fmt)
    assert os.path.exists(clust_opt.out_conf_name + "_2." + clust_opt.out_conf_fmt)
    assert os.path.exists(clust_opt.out_conf_name + "_3." + clust_opt.out_conf_fmt)



import pytest
from clusttraj.classify import classify_structures


def test_classify_structures(clust_opt, test_distmat, Z_matrix, clusters_seq):
    Z, clusters = classify_structures(clust_opt, test_distmat)

    assert Z == pytest.approx(Z_matrix, abs=1e-8)
    assert clusters == pytest.approx(clusters_seq, abs=1e-8)



"""Functions to plot the obtained results."""

from sklearn import manifold
from scipy.spatial.distance import squareform
import scipy.cluster.hierarchy as hcl
import matplotlib.pyplot as plt
import numpy as np
from .io import ClustOptions


def plot_clust_evo(clust_opt: ClustOptions, clusters: np.ndarray) -> None:
    """Plot the evolution of cluster classification over the given samples.

    Args:
        clust_opt (ClustOptions): The clustering options.
        clusters (np.ndarray): The cluster classifications for each sample.

    Returns:
        None
    """
    # plot evolution with o cluster in trajectory
    plt.figure(figsize=(25, 10))
    plt.plot(range(1, len(clusters) + 1), clusters, "o-", markersize=4)
    plt.xlabel("Sample Index")
    plt.ylabel("Cluster classification")
    plt.savefig(clust_opt.evo_name, bbox_inches="tight")


def plot_dendrogram(clust_opt: ClustOptions, Z: np.ndarray) -> None:
    """Plot a dendrogram based on hierarchical clustering.

    Parameters:
        clust_opt (ClustOptions): The options for clustering.
        Z (np.ndarray): The linkage matrix.

    Returns:
        None
    """
    # Plot the dendrogram
    plt.figure(figsize=(25, 10))
    plt.title("Hierarchical Clustering Dendrogram")
    plt.xlabel("Sample Index")
    plt.ylabel(r"RMSD ($\AA$)")

    hcl.dendrogram(
        Z,
        leaf_rotation=90.0,  # Rotates the x axis labels
        leaf_font_size=8.0,  # Font size for the x axis labels
    )

    # Add a horizontal line at the minimum RMSD value
    if clust_opt.silhouette_score:
        if isinstance(clust_opt.optimal_cut, (np.ndarray, list)):
            plt.axhline(clust_opt.optimal_cut[0], linestyle="--")
        elif isinstance(clust_opt.optimal_cut, (float, np.float32, np.float64)):
            plt.axhline(clust_opt.optimal_cut, linestyle="--")
        else:
            raise ValueError("optimal_cut must be a float or np.ndarray")
    else:
        plt.axhline(clust_opt.min_rmsd, linestyle="--")

    # Save the dendrogram to a file
    plt.savefig(clust_opt.dendrogram_name, bbox_inches="tight")


def plot_mds(clust_opt: ClustOptions, clusters: np.ndarray, distmat: np.ndarray) -> None:
    """Plot the multidimensional scaling (MDS) of the distance matrix.

    Args:
        clust_opt (ClustOptions): The clustering options.
        clusters (np.ndarray): The cluster labels.
        distmat (np.ndarray): The distance matrix.

    Returns:
        None
    """
    # Create a new figure
    plt.figure()

    # Initialize the MDS model
    mds = manifold.MDS(
        n_components=2,
        dissimilarity="precomputed",
        random_state=666,
        n_init=3,
        max_iter=200,
        eps=1e-3,
        n_jobs=clust_opt.n_workers,
        normalized_stress="auto",
    )

    # Perform MDS and get the 2D representation
    coords = mds.fit_transform(squareform(distmat))

    # Configure tick parameters
    plt.tick_params(
        axis="both",
        which="both",
        bottom=False,
        top=False,
        left=False,
        right=False,
        labelbottom=False,
        labelleft=False,
    )

    # Scatter plot the coordinates with cluster colors
    plt.scatter(
        coords[:, 0], coords[:, 1], marker="o", c=clusters, cmap=plt.cm.nipy_spectral
    )

    # Save the plot
    plt.savefig(clust_opt.mds_name, bbox_inches="tight")
